import torch
import torch.nn as nn
import numpy as np
from utils.shuffle import feature_shuffle_index


# ------------------------------- 6 expansion layers ------------------------------- #
class LinearExpansion(nn.Module):
    '''expansion 1D -> 2D'''

    def __init__(self, d_features, n_channel, n_depth):
        super().__init__()
        self.d_features = d_features
        self.n_channel = n_channel
        self.n_depth = n_depth

        self.expansion = nn.Linear(d_features, d_features * n_channel * n_depth)

    def forward(self, x):
        '''
            Arguments:
                x {Tensor, shape [batch, d_features]} -- input

            Returns:
                x {Tensor, shape [batch, n_channel * n_depth, d_features]} -- output
        '''
        x = self.expansion(x)
        x = x.view([-1, self.n_channel * self.n_depth, self.d_features])
        return x

    def initialize_param(self, init, *args):
        init(self.expansion.weight, *args)


class ReduceParamLinearExpansion(nn.Module):
    '''expansion 1D -> 2D'''

    def __init__(self, d_features, n_channel, n_depth):
        super().__init__()
        self.d_features = d_features
        self.n_channel = n_channel
        self.n_depth = n_depth

        d_hid = int(np.round(np.sqrt(d_features)))
        self.layer1 = nn.Linear(d_features, d_hid)
        self.layer2 = nn.Linear(d_hid, d_features * n_channel * n_depth)

    def forward(self, x):
        '''
            Arguments:
                x {Tensor, shape [batch, d_features]} -- input

            Returns:
                x {Tensor, shape [batch, n_channel * n_depth, d_features]} -- output
        '''
        x = self.layer1(x)
        x = self.layer2(x).view([-1, self.n_channel * self.n_depth, self.d_features])
        return x

    def initialize_param(self, init, *args):
        init(self.layer1.weight, *args)
        init(self.layer2.weight, *args)


class ConvExpansion(nn.Module):
    '''expansion 1D -> 2D'''

    def __init__(self, d_features, n_channel, n_depth):
        super().__init__()
        self.d_features = d_features
        self.n_channel = n_channel
        self.n_depth = n_depth

        self.conv = nn.Conv1d(1, n_channel * n_depth, kernel_size=3, padding=1)

    def forward(self, x):
        '''
            Arguments:
                x {Tensor, shape [batch, d_features]} -- input

            Returns:
                x {Tensor, shape [batch, n_channel * n_depth, d_features]} -- output
        '''
        assert x.dim() <= 3
        if x.dim() == 2:
            x = x.view(-1, 1, self.d_features)
        x = self.conv(x)
        return x

    def initialize_param(self, init, *args):
        init(self.conv.weight, *args)


class LinearConvExpansion(nn.Module):
    '''expansion 1D -> 2D'''

    def __init__(self, d_features, n_channel, n_depth):
        super().__init__()
        self.d_features = d_features
        self.n_channel = n_channel
        self.n_depth = n_depth

        self.d_hid = int(np.round(np.sqrt(d_features)))
        self.linear = nn.Linear(d_features, self.d_hid * d_features)
        self.conv = nn.Conv1d(self.d_hid, n_channel * n_depth, kernel_size=1)

    def forward(self, x):
        '''
            Arguments:
                x {Tensor, shape [batch, d_features]} -- input

            Returns:
                x {Tensor, shape [batch, n_channel * n_depth, d_features]} -- output
        '''
        x = self.linear(x).view(-1, self.d_hid, self.d_features)
        x = self.conv(x)
        return x

    def initialize_param(self, init, *args):
        init(self.linear.weight, *args)
        init(self.conv.weight, *args)


class ShuffleConvExpansion(nn.Module):
    '''expansion 1D -> 2D'''

    def __init__(self, d_features, n_channel, n_depth):
        super().__init__()
        self.d_features = d_features
        self.n_channel = n_channel
        self.n_depth = n_depth

        self.index = feature_shuffle_index(d_features, depth=self.dim)
        self.index = torch.tensor(self.index)
        self.d_features = d_features
        self.conv = nn.Conv1d(self.n_channel * self.n_depth, self.n_channel * self.n_depth, kernel_size=3, padding=1,
                              groups=self.n_channel * self.n_depth)

    def forward(self, x):
        '''
            Arguments:
                x {Tensor, shape [batch, d_features]} -- input

            Returns:
                x {Tensor, shape [batch, n_channel * n_depth, d_features]} -- output
        '''
        x = x[:, self.index]  # [batch, d_out]
        x = x.view(-1, self.n_channel * self.n_depth, self.d_features)  # [batch, n_channel * n_depth, d_features]
        x = self.conv(x)
        return x

    def initialize_param(self, init, *args):
        init(self.conv.weight, *args)


class ChannelWiseConv1d(nn.Module):
    def __init__(self, d_features, n_channel, n_depth):
        super().__init__()
        self.d_features = d_features
        self.n_channel = n_channel
        self.n_depth = n_depth

        self.conv = nn.Conv1d(n_depth, n_depth * n_channel, kernel_size=3, padding=1, groups=n_depth, bias=False)

    def forward(self, x):
        '''
            Arguments:
                x {Tensor, shape [batch, n_depth, d_features]} -- input

            Returns:
                x {Tensor, shape [batch, n_depth, n_channel, d_features]} -- output
        '''
        x = self.conv(x).view(
            [-1, self.n_depth, self.n_channel, self.d_features])  # [batch, n_depth, n_channel, d_features]
        return x

    def initialize_param(self, init, *args):
        init(self.conv.weight, *args)

class ChannelWiseConvExpansion(nn.Module):
    '''expansion 2D -> 3D -> flatten'''

    def __init__(self, d_features, n_channel, n_depth):
        super().__init__()
        self.d_features = d_features
        self.n_channel = n_channel
        self.n_depth = n_depth

        self.channel_conv = ChannelWiseConv1d(d_features, n_channel, n_depth)

    def forward(self, x):
        '''
            Arguments:
                x {Tensor, shape [batch, n_depth, d_features]

            Returns:
                x {Tensor, shape [batch, n_depth, n_channel * d_features]} -- output
        '''
        x = self.channel_conv(x)  # [batch, n_depth, n_channel, d_features]
        x = x.view([-1, self.n_depth, self.n_channel * self.d_features])  # [batch, n_depth, n_channel * d_features]
        return x

    def initialize_param(self, init, *args):
        self.channel_conv.initialize_param(init, *args)


class ChannelWiseConv2d(nn.Module):
    def __init__(self, d_features, n_channel, n_depth, c_total=3):
        super().__init__()
        self.d_features = d_features
        self.n_channel = n_channel
        self.n_depth = n_depth

        self.conv = nn.Conv2d(n_depth, n_depth * n_channel, kernel_size=(c_total, 3), padding=(0, 1), groups=n_depth, bias=False)

    def forward(self, x):
        '''
            Arguments:
                x {Tensor, shape [batch, c_total, n_depth, d_features]} -- input

            Returns:
                x {Tensor, shape [batch, n_depth, n_channel, d_features]} -- output
        '''
        x = x.permute([0, 2, 1, 3]).contiguous()  # [batch, n_depth, c_total, d_features]
        x = self.conv(x)  # [batch, n_depth * n_channel, 1, d_features]
        x = x.view(
            [-1, self.n_depth, self.n_channel, self.d_features])  # [batch, n_depth, n_channel, d_features]
        return x

    def initialize_param(self, init, *args):
        init(self.conv.weight, *args)

class ChannelWiseConvExpansionV2(nn.Module):
    '''expansion 3D -> 3D -> flatten'''

    def __init__(self, d_features, n_channel, n_depth, c_total=3):
        super().__init__()
        self.d_features = d_features
        self.n_channel = n_channel
        self.n_depth = n_depth

        self.channel_conv = ChannelWiseConv2d(d_features, n_channel, n_depth, c_total=c_total)

    def forward(self, x):
        '''
            Arguments:
                x {Tensor, shape [batch, c_total, n_depth, d_features]

            Returns:
                x {Tensor, shape [batch, n_depth, n_channel * d_features]} -- output
        '''
        x = self.channel_conv(x)  # [batch, n_depth, n_channel, d_features]
        x = x.view([-1, self.n_depth, self.n_channel * self.d_features])  # [batch, n_depth, n_channel * d_features]
        return x

    def initialize_param(self, init, *args):
        self.channel_conv.initialize_param(init, *args)



from framework.model import Model
from Modules.linear import LinearClassifier
from Modules.transformer import *
from Layers.transformer import *
from Layers.encodings import *
from torch.optim.adam import Adam
from Training.losses import *
from Training.evaluation import accuracy, precision_recall, Evaluator
from Training.control import TrainingControl, EarlyStopping
from tqdm import tqdm

def parse_data_enc(input_sequence, embedding):
    '''
    Returns:
        enc_output {Tensor, [batch_size, seq_length, d_v]} --
        non_pad_mask {Tensor, [n_head, seq_length, 1]} --
        slf_attn_mask {Tensor, [batch_size, seq_length, seq_length]} --
    '''

    slf_attn_mask = get_attn_key_pad_mask(seq_k=input_sequence, seq_q=input_sequence,
                                          padding_idx=0)  # [batch_size, seq_length, seq_length]
    non_pad_mask = get_non_pad_mask(input_sequence, padding_idx=0)  # [batch_size, seq_length, 1]

    embedding_sequence = embedding(input_sequence)

    return embedding_sequence, non_pad_mask, slf_attn_mask


def parse_data_dec(input_sequence, target_sequence, embedding):
    non_pad_mask = get_non_pad_mask(target_sequence)

    slf_attn_mask_subseq = get_subsequent_mask(target_sequence)
    slf_attn_mask_keypad = get_attn_key_pad_mask(seq_k=target_sequence, seq_q=target_sequence)
    slf_attn_mask = (slf_attn_mask_keypad + slf_attn_mask_subseq).gt(0)

    dec_enc_attn_mask = get_attn_key_pad_mask(seq_k=input_sequence, seq_q=target_sequence)

    embedding_sequence = embedding(target_sequence)
    return embedding_sequence, non_pad_mask, slf_attn_mask, dec_enc_attn_mask


class TransormerClassifierModel(Model):
    def __init__(
            self, save_path, log_path, d_features, d_meta, max_length, d_classifier, n_classes, threshold=None, embedding=None,
            stack='Encoder', position_encode='SinusoidPositionEncoding', optimizer=None, **kwargs):
        '''**kwargs: n_layers, n_head, dropout, use_bottleneck, d_bottleneck'''

        super().__init__(save_path, log_path)
        self.d_output = n_classes
        self.threshold = threshold
        self.max_length = max_length
        
        # ----------------------------- Model ------------------------------ #
        stack_dict = {
            'Encoder': Encoder,
            'Transformer': Transformer
        }
        encoding_dict = {
            'SinusoidPositionEncoding': SinusoidPositionEncoding,
            'LinearPositionEncoding': LinearPositionEncoding,
            'TimeFacilityEncoding': TimeFacilityEncoding
        }

        self.model = stack_dict[stack](encoding_dict[position_encode], d_features=d_features, max_seq_length=max_length, d_meta=d_meta, **kwargs)
        
        # --------------------------- Embedding  --------------------------- #
        if len(embedding) == 0:
            self.word_embedding = None
            self.USE_EMBEDDING = False

        else:
            self.word_embedding = nn.Embedding.from_pretrained(embedding)
            self.USE_EMBEDDING = True

        # --------------------------- Classifier --------------------------- #
        self.classifier = LinearClassifier(d_features * max_length, d_classifier, n_classes)
        
        # ------------------------------ CUDA ------------------------------ #
        # If GPU available, move the graph to GPU(s)
        self.CUDA_AVAILABLE = self.check_cuda()
        if self.CUDA_AVAILABLE:
            device_ids = list(range(torch.cuda.device_count()))
            self.model = nn.DataParallel(self.model, device_ids)
            self.classifier = nn.DataParallel(self.classifier, device_ids)
            self.model.to('cuda')
            self.classifier.to('cuda')
            assert (next(self.model.parameters()).is_cuda)
            assert (next(self.classifier.parameters()).is_cuda)
            pass

        else:
            print('CUDA not found or not enabled, use CPU instead')

        # ---------------------------- Optimizer --------------------------- #
        self.parameters = list(self.model.parameters()) + list(self.classifier.parameters())
        if optimizer == None:
            self.set_optimizer(Adam, lr=0.001, betas=(0.9, 0.999), weight_decay=0)

        # ------------------------ training control ------------------------ #
        self.controller = TrainingControl(max_step=100000, evaluate_every_nstep=100, print_every_nstep=10)
        self.early_stopping = EarlyStopping(patience=50)

        # --------------------- logging and tensorboard -------------------- #
        self.set_logger()
        self.set_summary_writer()
        # ---------------------------- END INIT ---------------------------- #

    def checkpoint(self, step):
        checkpoint = {
            'model_state_dict': self.model.state_dict(),
            'classifier_state_dict': self.classifier.state_dict(),
            'optimizer_state_dict': self.optimizer.state_dict(),
            'global_step': step}
        return checkpoint

    def train_epoch(self, train_dataloader, eval_dataloader, device, smothing, earlystop):
        ''' Epoch operation in training phase'''

        if device == 'cuda':
            assert self.CUDA_AVAILABLE
        # Set model and classifier training mode
        self.model.train()
        self.classifier.train()

        total_loss = 0
        batch_counter = 0

        # update param per batch
        for batch in tqdm(
                train_dataloader, mininterval=1,
                desc='  - (Training)   ', leave=False):  # training_data should be a iterable

            # get data from dataloader

            index, position, y = map(lambda x: x.to(device), batch)
            
            batch_size = len(index)

            input_feature_sequence, non_pad_mask, slf_attn_mask = parse_data_enc(index, self.word_embedding)

            # forward
            self.optimizer.zero_grad()
            logits, attn = self.model(input_feature_sequence, position, non_pad_mask, slf_attn_mask)
            logits = logits.view(batch_size, -1)
            logits = self.classifier(logits)

            # Judge if it's a regression problem
            if self.d_output == 1:
                pred = logits.sigmoid()
                loss = mse_loss(pred, y)

            else:
                pred = logits
                loss = cross_entropy_loss(pred, y, smoothing=smothing)

            # calculate gradients
            loss.backward()

            # update parameters
            self.optimizer.step()

            # get metrics for logging
            acc = accuracy(pred, y, threshold=self.threshold)
            precision, recall, precision_avg, recall_avg = precision_recall(pred, y, self.d_output,
                                                                            threshold=self.threshold)
            total_loss += loss.item()
            batch_counter += 1

            # training control
            state_dict = self.controller(batch_counter)

            if state_dict['step_to_print']:
                self.train_logger.info(
                    '[TRAINING]   - step: %5d, loss: %3.4f, acc: %1.4f, pre: %1.4f, rec: %1.4f' % (
                        state_dict['step'], loss, acc, precision[1], recall[1]))
                self.summary_writer.add_scalar('loss/train', loss, state_dict['step'])
                self.summary_writer.add_scalar('acc/train', acc, state_dict['step'])
                self.summary_writer.add_scalar('precision/train', precision[1], state_dict['step'])
                self.summary_writer.add_scalar('recall/train', recall[1], state_dict['step'])

            if state_dict['step_to_evaluate']:
                stop = self.val_epoch(eval_dataloader, device, state_dict['step'])
                state_dict['step_to_stop'] = stop

                if earlystop & stop:
                    break

            if self.controller.current_step == self.controller.max_step:
                state_dict['step_to_stop'] = True
                break

        return state_dict

    def val_epoch(self, dataloader, device, step=0, plot=False):
        ''' Epoch operation in evaluation phase '''
        if device == 'cuda':
            assert self.CUDA_AVAILABLE

        # Set model and classifier training mode
        self.model.eval()
        self.classifier.eval()

        # use evaluator to calculate the average performance
        evaluator = Evaluator()

        pred_list = []
        real_list = []

        with torch.no_grad():

            for batch in tqdm(
                    dataloader, mininterval=5,
                    desc='  - (Evaluation)   ', leave=False):  # training_data should be a iterable

                index, position, y = map(lambda x: x.to(device), batch)

                batch_size = len(index)

                input_feature_sequence, non_pad_mask, slf_attn_mask = parse_data_enc(index, self.word_embedding)

                # get logits
                logits, attn = self.model(input_feature_sequence, position, non_pad_mask, slf_attn_mask)
                logits = logits.view(batch_size, -1)
                logits = self.classifier(logits)

                if self.d_output == 1:
                    pred = logits.sigmoid()
                    loss = mse_loss(pred, y)

                else:
                    pred = logits
                    loss = cross_entropy_loss(pred, y, smoothing=False)

                acc = accuracy(pred, y, threshold=self.threshold)
                precision, recall, _, _ = precision_recall(pred, y, self.d_output, threshold=self.threshold)

                # feed the metrics in the evaluator
                evaluator(loss.item(), acc.item(), precision[1].item(), recall[1].item())

                '''append the results to the predict / real list for drawing ROC or PR curve.'''
            #     if plot:
            #         pred_list += pred.tolist()
            #         real_list += y.tolist()
            #
            # if plot:
            #     area, precisions, recalls, thresholds = pr(pred_list, real_list)
            #     plot_pr_curve(recalls, precisions, auc=area)

            # get evaluation results from the evaluator
            loss_avg, acc_avg, pre_avg, rec_avg = evaluator.avg_results()

            self.eval_logger.info(
                '[EVALUATION] - step: %5d, loss: %3.4f, acc: %1.4f, pre: %1.4f, rec: %1.4f' % (
                    step, loss_avg, acc_avg, pre_avg, rec_avg))
            self.summary_writer.add_scalar('loss/eval', loss_avg, step)
            self.summary_writer.add_scalar('acc/eval', acc_avg, step)
            self.summary_writer.add_scalar('precision/eval', pre_avg, step)
            self.summary_writer.add_scalar('recall/eval', rec_avg, step)


            state_dict = self.early_stopping(loss_avg)

            if state_dict['save']:
                checkpoint = self.checkpoint(step)
                self.save_model(checkpoint, self.save_path + '-step-%d_loss-%.5f'%(step,loss_avg))

            return state_dict['break']


    def train(self, max_epoch, lr, train_dataloader, eval_dataloader, device,
              smoothing=False, earlystop=False, save_mode='best'):
        assert save_mode in ['all', 'best']

        if not (lr is None):
            self.set_optimizer(AdamW, lr, betas=(0.9, 0.999), weight_decay=0.001)

        if self.USE_EMBEDDING:
            self.word_embedding = self.word_embedding.to(device)

        # train for n epoch
        for epoch_i in range(max_epoch):
            print('[ Epoch', epoch_i, ']')
            # set current epoch
            self.controller.set_epoch(epoch_i + 1)
            # train for on epoch
            state_dict = self.train_epoch(train_dataloader, eval_dataloader, device, smoothing, earlystop)

            # if state_dict['step_to_stop']:
            #     break

        checkpoint = self.checkpoint(state_dict['step'])

        self.save_model(checkpoint, self.save_path + '-step-%d' % state_dict['step'])

        self.train_logger.info('[INFO]: Finish Training, ends with %d epoch(s) and %d batches, in total %d training steps.' % (
            state_dict['epoch'] - 1, state_dict['batch'], state_dict['step']))

    def get_predictions(self, data_loader, device, max_batches=None, activation=None):

        if self.USE_EMBEDDING:
            self.word_embedding = self.word_embedding.to(device)

        pred_list = []
        real_list = []

        self.model.eval()
        self.classifier.eval()

        batch_counter = 0

        with torch.no_grad():
            for batch in tqdm(
                    data_loader,
                    desc='  - (Testing)   ', leave=False):

                index, position, y = map(lambda x: x.to(device), batch)

                input_feature_sequence, non_pad_mask, slf_attn_mask = parse_data_enc(index, self.word_embedding)

                # get logits
                logits, attn = self.model(input_feature_sequence, position, non_pad_mask, slf_attn_mask)
                logits = logits.view(logits.shape[0], -1)
                logits = self.classifier(logits)

                # Whether to apply activation function
                if activation != None:
                    pred = activation(logits)
                else:
                    pred = logits.softmax(dim=-1)
                pred_list += pred.tolist()
                real_list += y.tolist()

                if max_batches != None:
                    batch_counter += 1
                    if batch_counter >= max_batches:
                        break

        return pred_list, real_list





import torch
import torch.nn as nn
import numpy as np
from Layers.bottlenecks import LinearBottleneckLayer

class ReactionDotProduction(nn.Module):
    ''' Scaled Dot Productionss '''

    def __init__(self, temperature, attn_dropout=0.1):
        super().__init__()
        self.temperature = temperature
        self.dropout = nn.Dropout(attn_dropout)
        self.softmax = nn.Softmax(dim=2)

    def forward(self, query, key, value):
        '''
            Arguments:
                key {Tensor, shape [n_head * batch, d_features, n_depth_per_head]} -- expansion
                query {Tensor, shape [n_head * batch, 1, n_depth_per_head]} -- depth
                value {Tensor, shape [n_head * batch, 1, d_features]} -- value

            Returns:
                output {Tensor, shape [n_head * batch, 1, d_features]} -- output
                attn {Tensor, shape [n_head * batch, 1, d_features]} -- reaction attention
        '''
        attn = torch.bmm(query, key.transpose(1, 2))  # [n_head * batch, 1, d_features]
        # How should we set the temperature
        attn = attn / self.temperature

        attn = self.softmax(attn)  # softmax over d_f1
        attn = self.dropout(attn)
        output = torch.mul(attn, value)

        return output, attn


class ReactionAttentionLayerV1(nn.Module):
    '''Reaction Attention'''

    def __init__(self,expansion_layer, n_depth, d_features, d_meta, n_head, dropout,
                 use_bottleneck=True, d_bottleneck=None):
        super().__init__()

        self.d_features = d_features
        self.d_meta = d_meta
        self.n_head = n_head
        self.n_depth = n_depth
        self.use_bottleneck = use_bottleneck
        self.expansion = expansion_layer(d_features=d_features, n_channel=n_head, n_depth=n_depth) # output [batch, d_features, n_channel * n_depth]
        self.expansion.initialize_param(nn.init.xavier_normal_)

        # query, value map
        self.query = nn.Linear(d_meta, n_head * self.n_depth)
        self.value = nn.Linear(d_features, n_head * d_features)

        nn.init.xavier_normal_(self.query.weight)
        nn.init.xavier_normal_(self.value.weight)

        self.attention = ReactionDotProduction(temperature=np.power(self.n_depth, 0.5))

        self.layer_norm = nn.LayerNorm(d_features)

        self.fc = nn.Linear(n_head * d_features, d_features)
        nn.init.xavier_normal_(self.fc.weight)

        self.dropout = nn.Dropout(dropout)

        if use_bottleneck:
            self.bottleneck = LinearBottleneckLayer(d_features, d_bottleneck)

    def forward(self, features, meta):
        '''
            Arguments:
                feature_1 {Tensor, shape [batch, d_features]} -- feature part 1
                feature_2 {Tensor, shape [batch, d_meta]} -- feature part 2, can be categorical data

            Returns:
                output {Tensor, shape [batch, d_features]} -- output
                attn {Tensor, shape [n_head * batch, 1, d_features]} -- self attention
        '''
        d_features, d_meta, n_head, n_depth_per_head = self.d_features, self.d_meta, self.n_head, self.n_depth

        batch_size, _ = features.size()

        residual = features

        query = self.query(meta).view(batch_size, 1, n_head, n_depth_per_head)
        key = self.expansion(features).view(batch_size, n_depth_per_head, d_features, n_head)  # [batch, n_depth, n_head, d_features]
        value = self.value(features).view(batch_size, 1, n_head, d_features)
        # value = feature_1.repeat(1, n_head).view(batch_size, 1, n_head, d_features)

        query = query.permute(2, 0, 1, 3).contiguous().view(-1, 1, n_depth_per_head)
        key = key.permute(2, 0, 3, 1).contiguous().view(-1, d_features, n_depth_per_head)
        value = value.permute(2, 0, 1, 3).contiguous().view(-1, 1, d_features)

        output, attn = self.attention(query, key, value)

        output = output.view(n_head, batch_size, 1, d_features)
        output = output.permute(1, 2, 0, 3).contiguous().view(batch_size, -1)

        output = self.dropout(self.fc(output))
        output = self.layer_norm(output + residual)

        if self.use_bottleneck:
            output = self.bottleneck(output)

        return output, attn





import torch
import torch.nn as nn
import numpy as np
from Layers.reactionattention import *
from Layers.selfattention import *

class ReactionAttentionStack(nn.Module):
    ''' Reaction Attention Stack Module '''

    def __init__(
            self, expansion_layer, n_layers, n_head, d_features, d_meta, n_depth, d_bottleneck=256, dropout=0.1, mode='1d', use_bottleneck=True):
        super().__init__()
        self.layer_stack = nn.ModuleList([
            ReactionAttentionLayerV1(expansion_layer, n_depth, d_features, d_meta, n_head, dropout,
                                     use_bottleneck=use_bottleneck, d_bottleneck=d_bottleneck)
            for _ in range(n_layers)])

    def forward(self, features, meta_features):

        reaction_attn_list = []

        for ra_layer in self.layer_stack:
            features, reaction_attn = ra_layer(
                features, meta_features)
            reaction_attn_list += [reaction_attn]

        return features, reaction_attn_list


class SelfAttentionStack(nn.Module):
    ''' Self Attention Stack Module '''

    def __init__(
            self, expansion_layer, n_layers, n_head, d_f1, n_depth, d_hid=256, dropout=0.1):
        super().__init__()
        self.layer_stack = nn.ModuleList([
            SelfAttentionLayer(expansion_layer, n_head, n_depth, d_f1, d_hid=d_hid, dropout=dropout)
            for _ in range(n_layers)])

    def forward(self, feature_1, feature_2=None):

        self_attn_list = []

        for sa_layer in self.layer_stack:
            feature_1, self_attn = sa_layer(
                feature_1, feature_2)
            self_attn_list += [self_attn]

        return feature_1, self_attn_list


class AlternateStack(nn.Module):
    ''' Alternately stack the 2 attention blocks '''

    def __init__(
            self, expansion_layer, n_layers, n_head, d_f1, d_f2, n_depth, d_hid=256, dropout=0.1):
        super().__init__()
        self.layer_stack = nn.ModuleList([
            ReactionAttentionLayerV1(expansion_layer, n_head, n_depth, d_f1, d_f2, d_bottleneck=d_hid,
                                     dropout=dropout) if i % 2 == 0
            else SelfAttentionLayer(expansion_layer, n_head, n_depth, d_f1, d_hid=d_hid, dropout=dropout)
            for i in range(n_layers)])

    def forward(self, feature_1, feature_2):

        alternate_attn_list = []

        for attn_layer in self.layer_stack:
            feature_1, alternate_attn = attn_layer(
                feature_1, feature_2)
            alternate_attn_list += [alternate_attn]

        return feature_1, alternate_attn_list


class ParallelStack(nn.Module):
    ''' Stack the 2 attention blocks in parallel '''

    def __init__(
            self, expansion_layer, n_layers, n_head, d_f1, d_f2, n_depth, d_hid=256, dropout=0.1):
        super().__init__()
        self.n_layers = n_layers

        self.ra_stack = nn.ModuleList([
            ReactionAttentionLayerV1(expansion_layer, n_head, n_depth, d_f1, d_f2, dropout=dropout,
                                     use_bottleneck=False)
            for _ in range(n_layers)])

        self.sa_stack = nn.ModuleList([
            SelfAttentionLayer(expansion_layer, n_head, n_depth, d_f1, dropout=dropout, use_bottleneck=False)
            for _ in range(n_layers)])

        self.bottleneck_stack = nn.ModuleList([
            LinearBottleneckLayer(2 * d_f1, d_hid, d_out=d_f1)
            for _ in range(n_layers)])

    def forward(self, feature_1, feature_2):

        ensemble_attn_list = []

        for layer_index in range(self.n_layers):
            feature_1_ra, attn_ra = self.ra_stack[layer_index](feature_1, feature_2)
            feature_1_sa, attn_sa = self.sa_stack[layer_index](feature_1)
            feature_1 = torch.cat((feature_1_ra, feature_1_sa), dim=-1)
            feature_1 = self.bottleneck_stack(feature_1)
            ensemble_attn = torch.cat((attn_ra, attn_sa), dim=-2)
            ensemble_attn_list += [ensemble_attn]

        return feature_1, ensemble_attn_list


class ShuffleSelfAttentionStack(nn.Module):
    ''' Self Attention Stack Module '''

    def __init__(
            self,expansion_layer, n_depth, d_features, n_layers, n_head, n_channel, n_vchannel, d_bottleneck=256, dropout=0.1, mode='1d', use_bottleneck=True):
        super().__init__()
        self.layer_stack = nn.ModuleList([
            ShuffleSelfAttentionLayer(expansion_layer, n_depth, d_features, n_head, n_channel, n_vchannel, dropout=dropout,
                                      mode=mode, use_bottleneck=use_bottleneck, d_bottleneck=d_bottleneck)
            for _ in range(n_layers)])

    def forward(self, feature_1, feature_2=None):

        self_attn_list = []

        for sa_layer in self.layer_stack:
            feature_1, self_attn = sa_layer(
                feature_1)
            self_attn_list += [self_attn]

        return feature_1, self_attn_list

class ShuffleSelfAttentionStackV2(nn.Module):
    ''' Self Attention Stack Module '''

    def __init__(
            self,expansion_layer, n_depth, d_features, n_layers, n_head, n_channel, n_vchannel, d_bottleneck=256, dropout=0.1, mode='1d', use_bottleneck=True):
        super().__init__()
        self.layer_stack = nn.ModuleList([
            ShuffleSelfAttentionLayerV2(expansion_layer, n_depth, d_features, n_head, n_channel, n_vchannel, dropout=dropout,
                                      mode=mode, use_bottleneck=use_bottleneck, d_bottleneck=d_bottleneck)
            for _ in range(n_layers)])

    def forward(self, feature_1, feature_2=None):

        self_attn_list = []

        for sa_layer in self.layer_stack:
            feature_1, self_attn = sa_layer(
                feature_1)
            self_attn_list += [self_attn]

        return feature_1, self_attn_list



import torch
import torch.nn as nn
import numpy as np

def get_non_pad_mask(seq, padding_idx=0):
    '''
        Arguments:
            seq {batch, length} -- the word index of the sequence
            padding_idx -- padding symbol

        Returns:
            mask {Tensor, shape: [batch, length, 1]} -- non padding mask, 1 means not padding position while 0 means padding position
    '''
    assert seq.dim() == 2
    mask = seq.ne(padding_idx).type(torch.float)
    return mask.unsqueeze(-1)

def get_attn_key_pad_mask(seq_k, seq_q, padding_idx=0):
    '''
        For masking out the padding part of key sequence.
        Arguments:
            seq_k {Tensor, shape [batch, len_k]} -- key sequence
            seq_q {Tensor, shape [batch, len_q]} -- query sequence

        Returns:
            padding_mask {Tensor, shape [batch, len_q, len_k]} -- mask matrix, if the corresponding position is padding then True
            key mask [batch, k] -> expand q times [batch, len_q, len_k]

    '''

    # Expand to fit the shape of key query attention matrix.
    len_q = seq_q.size(1)
    padding_mask = seq_k.eq(padding_idx)
    padding_mask = padding_mask.unsqueeze(1).expand(-1, len_q, -1)  # b x lq x lk

    return padding_mask

def get_subsequent_mask(seq):
    ''' For masking out the subsequent info. '''

    sz_b, len_s = seq.size()
    subsequent_mask = torch.triu(
        torch.ones((len_s, len_s), device=seq.device, dtype=torch.uint8), diagonal=1)
    subsequent_mask = subsequent_mask.unsqueeze(0).expand(sz_b, -1, -1)  # b x ls x ls

    return subsequent_mask

class EncoderLayer(nn.Module):
    # def __init__(self, d_model, d_inner, n_head, d_k, d_v, dropout=0.1):
    def __init__(self, d_features, n_head, d_k, d_v, dropout,
                 use_bottleneck=False, d_bottleneck=128):
        super().__init__()

        # To reduce computation
        # d_reduce_param = np.floor(d_features / n_head).astype(int)

        self.self_attn = MultiHeadAttention(n_head, d_features, d_k, d_v, dropout)
        if use_bottleneck:
            self.bottleneck = PositionwiseFeedForward(d_features, d_bottleneck, dropout=dropout)

    def forward(self, enc_input, non_pad_mask=None, slf_attn_mask=None):
        '''
            Arguments:
                enc_input {Tensor, shape [batch, length, d_features]} -- input
                non_pad_mask {Tensor, shape [batch, length, 1]} -- index of which position in a sequence is a padding
                slf_attn_mask {Tensor, shape [batch, length, length]} -- self attention mask

            Returns:
                enc_output {Tensor, shape [batch, length, d_features]} -- output
                encoder_self_attn {n_head * batch, length, length} -- n_head self attention matrices
        '''

        enc_output, encoder_self_attn = self.self_attn(
            enc_input, enc_input, enc_input, mask=slf_attn_mask)
        enc_output *= non_pad_mask

        enc_output = self.bottleneck(enc_output)
        enc_output *= non_pad_mask

        return enc_output, encoder_self_attn


class DecoderLayer(nn.Module):
    ''' Compose with three layers '''

    # def __init__(self, d_model, d_inner, n_head, d_k, d_v, dropout=0.1):
    def __init__(self, d_features, n_head, dropout,
                 use_bottleneck=False, d_bottleneck=128):
        super().__init__()

        # To reduce computation
        d_reduce_param = np.floor(d_features / n_head).astype(int)

        self.slf_attn = MultiHeadAttention(n_head, d_features, d_reduce_param, d_reduce_param, dropout=dropout)
        self.enc_attn = MultiHeadAttention(n_head, d_features, d_reduce_param, d_reduce_param, dropout=dropout)
        if use_bottleneck:
            self.bottleneck = PositionwiseFeedForward(d_features, d_bottleneck, dropout=dropout)

    def forward(self, dec_input, enc_output, non_pad_mask=None, slf_attn_mask=None, dec_enc_attn_mask=None):
        '''
            Arguments:
                dec_input {Tensor, shape [batch, in_length, d_features]} -- decoder input
                enc_output {Tensor, shape [batch, tg_length, d_features]} -- encoder output
                non_pad_mask {Tensor, shape [batch, tg_length, 1]} -- index of which position in a sequence is a padding
                slf_attn_mask {Tensor, shape [batch, tg_length, tg_length]} -- self attention mask
                dec_enc_attn_mask {Tensor, shape [batch, tg_length, in_length]}
            Returns:
                dec_output {Tensor, shape [batch_size, seq_length, d_features]} -- output
                dec_slf_attn {Tensor, shape [n_head, seq_length, 1]} -- decoder self attention
                dec_enc_attn {Tensor, shape [batch_size, seq_length, seq_length]} -- decoder-encoder attention
        '''

        dec_output, dec_slf_attn = self.slf_attn(
            dec_input, dec_input, dec_input, mask=slf_attn_mask)
        dec_output *= non_pad_mask

        dec_output, dec_enc_attn = self.enc_attn(
            dec_output, enc_output, enc_output, mask=dec_enc_attn_mask)
        dec_output *= non_pad_mask

        dec_output = self.pos_ffn(dec_output) # wider the network
        dec_output *= non_pad_mask

        return dec_output, dec_slf_attn, dec_enc_attn


class ScaledDotProductAttention(nn.Module):
    ''' Scaled Dot-Product Attention '''

    def __init__(self, temperature, attn_dropout=0.1):
        super().__init__()
        self.temperature = temperature
        self.dropout = nn.Dropout(attn_dropout)
        self.softmax = nn.Softmax(dim=2)

    def forward(self, query, key, value, mask=None):
        '''
            Arguments:
                query {Tensor, shape [n_head * batch, q_length, dk]} -- query
                key {Tensor, shape [n_head * batch, k_length, dk]} -- key
                value {Tensor, shape [n_head * batch, v_length, dv]} -- value
                mask {Tensor, shape [n_head * batch, q_length, k_length]} --self attn mask

            Returns:
                output {Tensor, shape [n_head * batch, q_length, dv] -- output
                attn {Tensor, shape [n_head * batch, q_length, k_length] -- self attention

        '''
        attn = torch.bmm(query, key.transpose(1, 2)) # [n_head * batch, q_length, k_length]
        attn = attn / self.temperature

        if mask is not None:
            attn = attn.masked_fill(mask, -np.inf)

        attn = self.softmax(attn) # softmax over k_length
        attn = self.dropout(attn)
        output = torch.bmm(attn, value)

        return output, attn


class MultiHeadAttention(nn.Module):
    def __init__(self, n_head, d_features, d_k, d_v, dropout=0.1):
        super().__init__()
        self.n_head = n_head
        self.d_k = d_k
        self.d_v = d_v

        self.w_qs = nn.Linear(d_features, n_head * d_k)
        self.w_ks = nn.Linear(d_features, n_head * d_k)
        self.w_vs = nn.Linear(d_features, n_head * d_v)
        nn.init.normal_(self.w_qs.weight, mean=0, std=np.sqrt(2.0 / (d_features + d_k)))
        nn.init.normal_(self.w_ks.weight, mean=0, std=np.sqrt(2.0 / (d_features + d_k)))
        nn.init.normal_(self.w_vs.weight, mean=0, std=np.sqrt(2.0 / (d_features + d_v)))

        self.attention = ScaledDotProductAttention(temperature=np.power(d_k, 0.5))
        self.layer_norm = nn.LayerNorm(d_features)

        self.fc = nn.Linear(n_head * d_v, d_features)
        nn.init.xavier_normal_(self.fc.weight)

        self.dropout = nn.Dropout(dropout)

    def forward(self, query, key, value, mask=None):
        '''
            Arguments:
                query {Tensor, shape [batch, q_length, d_features]} -- query
                key {Tensor, shape [batch, k_length, d_features]} -- key
                value {Tensor, shape [batch, v_length, d_features]} -- value
                mask {Tensor, shape [batch, q_length, k_length]} --self attn mask

            Returns:
                output {Tensor, shape [batch, q_length, d_features]} -- output
                attn {Tensor, shape [n_head * batch, q_length, k_length] -- self attention
        '''
        d_k, d_v, n_head = self.d_k, self.d_v, self.n_head

        sz_b, len_q, _ = query.size()
        sz_b, len_k, _ = key.size()
        sz_b, len_v, _ = value.size()

        assert len_k == len_v

        residual = query

        query = self.w_qs(query).view(sz_b, len_q, n_head, d_k)  # target
        # [batch_size, seq_length, w2v_length] -> [batch_size, seq_length, n_head * dk] -> [batch_size, seq_length, n_head, dk]
        key = self.w_ks(key).view(sz_b, len_k, n_head, d_k)
        value = self.w_vs(value).view(sz_b, len_v, n_head, d_v)

        query = query.permute(2, 0, 1, 3).contiguous().view(-1, len_q, d_k)  # [n_head * batch_size, seq_length, dk]
        key = key.permute(2, 0, 1, 3).contiguous().view(-1, len_k, d_k)  # (n*b) x lk x dk
        value = value.permute(2, 0, 1, 3).contiguous().view(-1, len_v, d_v)  # (n*b) x lv x dv

        if len(mask) != 0:
            mask = mask.repeat(n_head, 1, 1)  # [n_head * batch_size, seq_length, seq_length]

        output, attn = self.attention(query, key, value, mask=mask)

        output = output.view(n_head, sz_b, len_q, d_v)
        output = output.permute(1, 2, 0, 3).contiguous().view(sz_b, len_q, -1)  # b x lq x (n*dv)

        output = self.dropout(self.fc(output))
        output = self.layer_norm(output + residual)

        return output, attn


class PositionwiseFeedForward(nn.Module):
    ''' A two-1x1 Conv-layer bottleneck module '''

    def __init__(self, d_in, d_hid, dropout=0.1):
        super().__init__()
        self.w_1 = nn.Conv1d(d_in, d_hid, 1)  # position-wise
        self.w_2 = nn.Conv1d(d_hid, d_in, 1)  # position-wise
        self.layer_norm = nn.LayerNorm(d_in)
        self.dropout = nn.Dropout(dropout)

    def forward(self, x):
        '''
            Arguments:
                x {Tensor, shape [batch_size, length, d_features]}

            Returns:
                x {Tensor, shape [batch_size, length, d_features]}

        '''
        residual = x
        output = x.transpose(1, 2)
        output = nn.functional.relu(self.w_1(output))
        output = self.w_2(output)
        output = output.transpose(1, 2)
        output = self.dropout(output)
        output = self.layer_norm(output + residual)
        return output


import torch
import torch.nn.functional as F
from utils.embeddings import one_hot_embedding

def cross_entropy_loss(logits, real, smoothing=False):
    ''' Calculate cross entropy loss, apply label smoothing if needed. '''

    real = real.contiguous().view(-1).long()

    if smoothing:
        eps = 0.1
        n_class = logits.size(1)

        one_hot = torch.zeros_like(logits).scatter(1, real.view(-1, 1), 1)
        one_hot = one_hot * (1 - eps) + (1 - one_hot) * eps / (n_class - 1)
        log_prb = F.log_softmax(logits, dim=1)

        non_pad_mask = real.ne(0)
        loss = -(one_hot * log_prb).sum(dim=1)
        loss = loss.masked_select(non_pad_mask).sum()  # average later
    else:
        # loss = F.cross_entropy(logits, real, ignore_index=0, reduction='sum')
        loss = F.cross_entropy(logits, real)

    return loss


def mse_loss(pred, real):
    loss = F.mse_loss(pred, real)

    return loss


import torch
import torch.nn as nn
from Layers.normalization import LayerNorm
from Layers.transformer import EncoderLayer

class EncoderDecoder(nn.Module):
    """
    A standard Encoder-Decoder architecture. Base for this and many
    other models.
    """

    def __init__(self, encoder, decoder, src_embed, tgt_embed, generator):
        super().__init__()
        self.encoder = encoder
        self.decoder = decoder
        self.src_embed = src_embed
        self.tgt_embed = tgt_embed
        self.generator = generator

    def forward(self, src, tgt, src_mask, tgt_mask):
        "Take in and process masked src and target sequences."
        return self.decode(self.encode(src, src_mask), src_mask,
                           tgt, tgt_mask)

    def encode(self, src, src_mask):
        return self.encoder(self.src_embed(src), src_mask)

    def decode(self, memory, src_mask, tgt, tgt_mask):
        return self.decoder(self.tgt_embed(tgt), memory, src_mask, tgt_mask)


class Encoder(nn.Module):
    "Core encoder is a stack of N layers"

    def __init__(self, n_layers, input_shape, output_shape, dropout):
        super().__init__()
        self.layer_stack = nn.ModuleList([EncoderLayer(input_shape, output_shape, dropout=dropout)
                                          for _ in range(n_layers)])
        self.layer_norm = LayerNorm(input_shape)

    def forward(self, x, mask):
        "Pass the input (and mask) through each layer in turn."
        for layer in self.layer_stack:
            x = layer(x, mask)
        return self.layer_norm(x)



import torch
import torch.nn as nn
import numpy as np

class PlainConvLayer(nn.Module):
    def __init__(self, n_channels, d_features, dropout):
        super().__init__()

        self.conv1 = nn.Conv1d(1, n_channels, kernel_size=3, bias=False)
        self.conv2 = nn.Conv1d(n_channels, 1, kernel_size=3, bias=False)
        nn.init.xavier_normal(self.conv1.weight)
        nn.init.xavier_normal(self.conv2.weight)

        self.activation = nn.ReLU()
        self.dropout = nn.Dropout(dropout)
        self.layer_norm = nn.LayerNorm(d_features)


    def forward(self, x):
        '''
            Arguments:
                x {Tensor, shape [batch, 1, d_features]}
            Returns:
                output {Tensor, shape [batch, 1, d_features]}
        '''
        residual = x
        output = self.conv1(x)  # [batch, n_channels, d_features]
        output = self.activation(output)
        output = self.layer_norm(output)
        output = self.conv2(output)
        output = self.activation(output)
        output = self.layer_norm(output + residual)

        return output







import pytorch_lightning as pl


# class ReactionModelLightning(pl.LightningModule):
#     def __init__(self, dataloader, stack, d_reactant, d_bottleneck, d_classifier, d_output, threshold=None,
#                  n_layers=6, n_head=8, dropout=0.1):
#         super().__init__()
#
#         self.dataloader = dataloader
#         feature1_dim, feature2_dim = self.dataloader.get_feature_dim()
#         self.model = stack(n_layers, n_head, feature1_dim, feature2_dim, d_reactant, d_bottleneck,
#                            dropout)
#         self.fc1 = nn.Linear(feature1_dim, d_classifier)
#         self.fc2 = nn.Linear(d_classifier, d_output)
#
#         self.d_output = d_output
#         self.threshold = threshold
#
#     def forward(self, feature_1, feature_2):
#         output = self.model(feature_1, feature_2)
#         output = self.fc1(output[0])
#         output = nn.functional.relu(output)
#         output = self.fc2(output)
#         return output
#
#     def training_step(self, batch, batch_nb):
#         feature_1, feature_2, y = batch
#         logits = self.forward(feature_1, feature_2)
#         pred = logits.sigmoid()
#
#         if y.shape[-1] == 1:
#             loss = mse_loss(pred, y)
#
#         else:
#             loss = cross_entropy_loss(pred, y, smoothing=True)
#
#         acc = accuracy(pred, y, threshold=self.threshold)
#         _, _, precision_avg, recall_avg = precision_recall(pred, y, self.d_output, threshold=self.threshold)
#         tensorboard_logs = {'train_loss': loss, 'train_acc': acc,
#                             'train_precision': precision_avg,
#                             'train_recall': recall_avg}
#
#         return {'loss': loss, 'acc': acc, 'precision': precision_avg, 'recall': recall_avg, 'log': tensorboard_logs}
#
#     def validation_step(self, batch, batch_nb):
#         feature_1, feature_2, y = batch
#         logits = self.forward(feature_1, feature_2)
#         pred = logits.sigmoid()
#
#         if y.shape[-1] == 1:
#             loss = mse_loss(pred, y)
#
#         else:
#             loss = cross_entropy_loss(pred, y, smoothing=False)
#
#         acc = accuracy(pred, y, threshold=self.threshold)
#         _, _, precision_avg, recall_avg = precision_recall(pred, y, self.d_output, threshold=self.threshold)
#
#         return {'val_loss': loss, 'val_acc': acc,
#                 'val_precision': precision_avg,
#                 'val_recall': recall_avg}
#
#     def validation_end(self, outputs):
#         avg_loss = torch.stack([x['val_loss'] for x in outputs]).mean()
#         avg_acc = torch.stack([x['val_acc'] for x in outputs]).mean()
#         avg_precision = torch.stack([x['val_precision'] for x in outputs]).mean()
#         avg_recall = torch.stack([x['val_recall'] for x in outputs]).mean()
#
#         tensorboard_logs = {'val_loss': avg_loss, 'val_acc': avg_acc,
#                             'val_precision': avg_precision,
#                             'val_recall': avg_recall}
#
#         return {'loss': avg_loss, 'acc': avg_acc, 'precision': avg_precision, 'recall': avg_recall,
#                 'log': tensorboard_logs}
#
#     def configure_optimizers(self):
#         # return AdamOptimizer(self.parameters(), 1000, 0.01)
#         return Adam(self.parameters(), lr=0.005, betas=(0.9, 0.999))
#
#     @pl.data_loader
#     def train_dataloader(self):
#         return self.dataloader.train_dataloader()
#
#     @pl.data_loader
#     def val_dataloader(self):
#         return self.dataloader.val_dataloader()
#
#     @pl.data_loader
#     def test_dataloader(self):
#         return self.dataloader.test_dataloader()


import numpy as np


class TrainingControl():
    def __init__(self, max_step, evaluate_every_nstep, print_every_nstep):
        self.state_dict = {
            'epoch': 0,
            'batch': 0,
            'step': 0,
            'step_to_evaluate': False,
            'step_to_print': False,
            'step_to_stop': False
        }
        self.max_step = max_step
        self.eval_every_n = evaluate_every_nstep
        self.print_every_n = print_every_nstep
        self.current_epoch = 0
        self.current_batch = 0
        self.current_step = 0

    def __call__(self, batch):
        self.current_step += 1
        self.state_dict['batch'] = batch
        self.state_dict['step'] = self.current_step
        self.state_dict['step_to_evaluate'] = np.equal(np.mod(self.current_step, self.eval_every_n), 0)
        self.state_dict['step_to_print'] = np.equal(np.mod(self.current_step, self.print_every_n), 0)
        self.state_dict['step_to_stop'] = np.equal(self.current_step, self.max_step)
        return self.state_dict

    def set_epoch(self, epoch):
        self.state_dict['epoch'] = epoch

    def reset_state(self):
        self.state_dict = {
            'epoch': 0,
            'batch': 0,
            'step': 0,
            'step_to_evaluate': False,
            'step_to_print': False,
            'step_to_stop': False
        }
        self.current_epoch = 0
        self.current_batch = 0
        self.current_step = 0


class EarlyStopping():
    def __init__(self, patience, mode='best'):
        self.patience = patience
        self.mode = mode
        self.best_loss = 9999
        self.waitting = 0
        self.state_dict = {
            'save': False,
            'break': False
        }

    def __call__(self, val_loss):
        self.state_dict['save'] = False
        self.state_dict['break'] = False

        if val_loss <= self.best_loss:
            self.best_loss = val_loss
            self.waitting = 0
            self.state_dict['save'] = True

        else:
            self.waitting += 1

            if self.mode == 'best':
                self.state_dict['save'] = False
            else:
                self.state_dict['save'] = True

            if self.waitting == self.patience:
                self.state_dict['break'] = True

        return self.state_dict
    def reset_state(self):
        self.best_loss = 9999
        self.waitting = 0
        self.state_dict = {
            'save': False,
            'break': False
        }




from abc import abstractmethod
from utils.loggings import logger
import torch
from torch.utils.tensorboard import SummaryWriter

class Model():
    def __init__(self, save_path, log_path):
        self.save_path = save_path
        self.log_path = log_path
        self.model = None
        self.classifier = None
        self.parameters = None
        self.optimizer = None
        self.train_logger = None
        self.eval_logger = None
        self.summary_writer = None

    @abstractmethod
    def loss(self, **kwargs):
        pass

    @abstractmethod
    def train(self, **kwargs):
        pass

    @abstractmethod
    def evaluate(self, **kwargs):
        pass

    @abstractmethod
    def checkpoint(self, **kwargs):
        pass

    def set_optimizer(self, Optimizer, lr, **kwargs):
        self.optimizer = Optimizer(self.parameters, lr=lr, **kwargs)

    def set_logger(self, mode='a'):
        self.train_logger = logger('train',self.log_path + '-train', mode=mode)
        self.eval_logger = logger('eval', self.log_path + '-eval', mode=mode)

    def set_summary_writer(self):
        self.summary_writer = SummaryWriter(self.log_path + 'tensorboard')

    def check_cuda(self):
        if torch.cuda.is_available():
            print("INFO: CUDA device exists")
            return torch.cuda.is_available()


    def resume_checkpoint(self, checkpoint_path):
        checkpoint = torch.load(checkpoint_path)

        if self.model != None:
            model_state_dict = checkpoint['model_state_dict']
            self.model.load_state_dict(model_state_dict)
            self.model.train()

        classifier_state_dict = checkpoint['classifier_state_dict']
        self.classifier.load_state_dict(classifier_state_dict)
        self.classifier.train()

    def save_model(self, checkpoint, save_path):
        torch.save(checkpoint, save_path)

    def load_model(self, checkpoint_path):
        checkpoint = torch.load(checkpoint_path)

        if self.model != None:
            model_state_dict = checkpoint['model_state_dict']
            self.model.load_state_dict(model_state_dict)
            self.model.eval()

        classifier_state_dict = checkpoint['classifier_state_dict']
        self.classifier.load_state_dict(classifier_state_dict)
        self.classifier.eval()

    def count_parameters(self):
        try:
            assert self.model != None
            model_params = sum(p.numel() for p in self.model.parameters() if p.requires_grad)
            print('Number of Model Parameters: %d'%model_params)
        except:
            print('No Model specified')

        try:
            assert self.classifier != None
            classifier = sum(p.numel() for p in self.classifier.parameters() if p.requires_grad)
            print('Number of Model Classifier: %d' % classifier)
        except:
            print('No Classifier specified')





import torch
import torch.nn as nn

class LayerNorm(nn.Module):
    def __init__(self, input_shape, eps=1e-6):
        super().__init__()
        self.gamma = nn.Parameter(torch.ones(input_shape))
        self.beta = nn.Parameter(torch.zeros(input_shape))
        self.eps = eps

    def forward(self, x):
        mean = x.mean(-1, keepdim=True)
        std = x.std(-1, keepdim=True)
        return self.gamma * (x - mean) / (std + self.eps) + self.beta


import torch
import torch.nn as nn
import torch.nn.functional as F
import numpy as np
from Layers.expansions import feature_shuffle_index
from Layers.bottlenecks import *


class ScaledDotProduction(nn.Module):
    '''Scaled Dot Production'''

    def __init__(self, temperature, attn_dropout=0.1):
        super().__init__()
        self.temperature = temperature
        self.dropout = nn.Dropout(attn_dropout)
        self.softmax = nn.Softmax(dim=2)

    def forward(self, query, key, value):
        '''
            Arguments:
                query {Tensor, shape [n_head * batch, n_depth, n_channel * d_features]} -- query
                key {Tensor, shape [n_head * batch, n_depth, n_channel * d_features]} -- key
                value {Tensor, shape [n_head * batch, n_depth, n_vchannel * d_features]} -- value

            Returns:
                output {Tensor, shape [n_head * batch, n_depth, n_vchannel * d_features] -- output
                attn {Tensor, shape [n_head * batch, n_depth, n_depth] -- reaction attention
        '''
        attn = torch.bmm(query, key.transpose(2, 1))  # [n_head * batch, n_depth, n_depth]
        # How should we set the temperature
        attn = attn / self.temperature

        attn = self.softmax(attn)  # softmax over d_f1
        attn = self.dropout(attn)
        output = torch.bmm(attn, value)

        return output, attn


class SelfAttentionLayer(nn.Module):
    '''Self Attention'''

    def __init__(self, expansion_layer, n_depth, d_features, n_head, d_hid=256, dropout=0.1, use_bottleneck=True):
        super().__init__()
        self.d_features = d_features
        self.n_head = n_head
        self.n_depth = n_depth
        self.use_bottleneck = use_bottleneck

        self.query = expansion_layer(d_features=d_features, n_channel=n_head, n_depth=n_depth)
        self.key = expansion_layer(d_features=d_features, n_channel=n_head, n_depth=n_depth)
        self.value = expansion_layer(d_features=d_features, n_channel=n_head, n_depth=1)

        self.query.initialize_param(nn.init.xavier_normal_)
        self.key.initialize_param(nn.init.xavier_normal_)
        self.value.initialize_param(nn.init.xavier_normal_)

        self.attention = ScaledDotProduction(temperature=np.power(n_depth, 0.5))

        self.layer_norm = nn.LayerNorm(d_features)

        self.fc = nn.Linear(n_head * d_features, d_features)
        nn.init.xavier_normal(self.fc.weight)

        self.dropout = nn.Dropout(dropout)

        if use_bottleneck:
            self.bottleneck = LinearBottleneckLayer(d_features, d_hid)

    def forward(self, feature_1, feature_2=None):
        '''
            Arguments:
                feature_1 {Tensor, shape [batch, d_features]} -- feature part 1

            Returns:
                output {Tensor, shape [batch, d_features]} -- output
                attn {Tensor, shape [n_head * batch, d_features, d_features]} -- self attention
        '''
        d_features, n_head, n_depth, n_vchannel = self.d_features, self.n_head, self.n_depth, self.n_v

        batch_size, _ = feature_1.size()

        residual = feature_1

        query = self.query(feature_1).view(batch_size, d_features, n_head, n_depth)
        key = self.key(feature_1).view(batch_size, d_features, n_head, n_depth)
        value = self.value(feature_1).view(batch_size, 1, n_head, d_features)

        query = query.permute(2, 0, 1, 3).contiguous().view(-1, d_features, n_depth)
        key = key.premute(2, 0, 1, 3).contiguous().view(-1, d_features, n_depth)
        value = value.permute(2, 0, 1, 3).contiguous().view(-1, 1, d_features)

        output, attn = self.attention(query, key, value)

        output = output.view(n_head, batch_size, d_features, 1)
        output = output.permute(1, 2, 0, 3).contiguous().view(batch_size, -1)

        output = self.dropout(self.fc(output))
        output = self.layer_norm(output + residual)

        if self.use_bottleneck:
            output = self.bottleneck(output)

        return output, attn


class ShuffleSelfAttention(nn.Module):
    '''Self Attention'''

    def __init__(self, expansion_layer, n_depth, d_features, n_head, n_channel, n_vchannel):
        super().__init__()
        self.n_depth = n_depth
        self.d_features = d_features
        self.n_head = n_head
        self.n_channel = n_channel
        self.n_vchannel = n_vchannel
        total_channels = n_head * n_channel
        total_vchannels = n_head * n_vchannel

        self.query = expansion_layer(d_features=d_features, n_channel=total_channels, n_depth=n_depth)
        self.key = expansion_layer(d_features=d_features, n_channel=total_channels, n_depth=n_depth)
        self.value = expansion_layer(d_features=d_features, n_channel=total_vchannels, n_depth=n_depth)

        self.query.initialize_param(nn.init.xavier_normal_)
        self.key.initialize_param(nn.init.xavier_normal_)
        self.value.initialize_param(nn.init.xavier_normal_)

        self.attention = ScaledDotProduction(temperature=np.power(n_depth, 0.5))

    def forward(self, feature_map):
        '''
            Arguments:
                feature_map {Tensor, shape [batch, n_depth, d_features]} -- feature part 1

            Returns:
                output {Tensor, shape [batch, n_head * n_vchannel, n_depth, d_features]} -- output
                attn {Tensor, shape [n_head * batch, n_depth, n_depth]} -- self attention
        '''
        d_f1, n_head, n_channel, n_vchannel, n_depth = self.d_features, self.n_head, self.n_channel, self.n_vchannel, self.n_depth

        batch_size, _, _ = feature_map.size()

        query = self.query(
            feature_map)  # shape [batch, n_depth, n_head * n_channel * d_features] for using ChannelWiseConvExpansion, otherwise [batch, d_features, n_head * n_channel * n_depth]
        dim_2nd = query.shape[1]  # n_depth or d_features
        query = query.view(batch_size, dim_2nd, n_head,
                           -1)  # [batch, n_depth, n_head, n_channel * d_features] or [batch, d_features, n_head, n_channel * n_depth]
        key = self.key(feature_map)
        key = key.view(batch_size, dim_2nd, n_head, -1)
        value = self.value(feature_map)
        value = value.view(batch_size, dim_2nd, n_head, -1)

        query = query.permute(2, 0, 1, 3)  # [n_head, batch, n_depth, n_channel * d_features]
        query = query.contiguous().view(n_head * batch_size, dim_2nd,
                                        -1)  # [n_head * batch, n_depth, n_channel * d_features] or [n_head * batch, d_features, n_channel * n_depth]
        key = key.permute(2, 0, 1, 3).contiguous().view(n_head * batch_size, dim_2nd, -1)
        value = value.permute(2, 0, 1, 3).contiguous().view(n_head * batch_size, dim_2nd, -1)

        output, attn = self.attention(query, key,
                                      value)  # [n_head * batch, n_depth, n_vchannel * d_features]

        output = output.view(n_head, batch_size, dim_2nd,
                             -1)  # [n_head, batch, n_depth, n_vchannel * d_features]
        output = output.permute(1, 2, 0, 3).contiguous().view(batch_size, dim_2nd, n_head * n_vchannel,
                                                              -1)  # [batch, n_depth, n_head * n_vchannel, d_features]
        output = output.transpose(1, 2)  # [batch, n_head * n_vchannel, n_depth, d_features]

        return output, attn


class ShuffleSelfAttentionLayer(nn.Module):
    def __init__(self, expansion_layer, n_depth, d_features, n_head, n_channel, n_vchannel, dropout,
                 mode='1d', use_bottleneck=True, d_bottleneck=None):
        super().__init__()

        assert n_depth > 1

        self.d_features = d_features
        self.n_depth = n_depth
        self.mode = mode
        self.use_bottleneck = use_bottleneck

        self.index = feature_shuffle_index(d_features, depth=n_depth)
        self.index = torch.tensor(self.index)

        self.shuffle_slf_attn = ShuffleSelfAttention(expansion_layer, n_depth, d_features, n_head, n_channel,
                                                     n_vchannel)

        self.layer_norm = nn.LayerNorm(d_features)
        self.dropout = nn.Dropout(dropout)

        if mode == '1d':
            self.conv = nn.Conv2d(n_head * n_vchannel, 1, kernel_size=(n_depth, 1), bias=False)
            nn.init.xavier_normal(self.conv.weight)

        elif mode == '2d':
            self.conv = nn.Conv2d(n_head * n_vchannel, 1, kernel_size=(1, 1), bias=False)  # or use fc
            nn.init.xavier_normal(self.conv.weight)
        else:
            pass

        if use_bottleneck:
            self.bottleneck = ShuffleBottleneckLayer(n_depth, d_features, mode, d_hid=d_bottleneck)

    def forward(self, features):
        if features.dim() == 2:
            feature_map = features[:, self.index]  # [batch, n_depth * d_features]
            feature_map = feature_map.view([-1, self.n_depth, self.d_features])  # [batch, n_depth, d_features]
        else:
            feature_map = features

        residual = features

        output, attn = self.shuffle_slf_attn(
            feature_map)  # output shape [batch, n_head * n_vchannel, n_depth, d_features]
        output = self.conv(output)

        if self.mode == '1d':
            if features.dim() == 3:
                residual = 0
                print('[WARNING]: Got features with dimension=3 while output mode is 1d, cannot apply residual')
            output = output.view([-1, self.d_features])

        elif self.mode == '2d':
            residual = feature_map
            output = output.view([-1, self.n_depth, self.d_features])  # [batch, n_depth, d_features]

        output = self.dropout(output)
        output = self.layer_norm(output + residual)

        if self.use_bottleneck:
            output = self.bottleneck(output)

        return output, attn


class ShuffleSelfAttentionV2(nn.Module):
    '''Self Attention'''

    def __init__(self, expansion_layer, n_depth, d_features, n_head, n_channel, n_vchannel, d_block, c_total=3):
        super().__init__()
        self.n_depth = n_depth
        self.d_features = d_features
        self.n_head = n_head
        self.n_channel = n_channel
        self.n_vchannel = n_vchannel
        self.d_block = d_block
        self.c_total = c_total
        total_channels = n_head * n_channel
        total_vchannels = n_head * n_vchannel

        self.query = expansion_layer(d_features=d_features, n_channel=total_channels, n_depth=n_depth)
        self.key = expansion_layer(d_features=d_features, n_channel=total_channels, n_depth=n_depth)
        self.value = expansion_layer(d_features=d_features, n_channel=total_vchannels, n_depth=n_depth)

        self.query.initialize_param(nn.init.xavier_normal_)
        self.key.initialize_param(nn.init.xavier_normal_)
        self.value.initialize_param(nn.init.xavier_normal_)

        self.attention = ScaledDotProduction(temperature=np.power(n_depth, 0.5))

    def forward(self, feature_map):
        '''
            Arguments:
                feature_map {Tensor, shape [batch, c_total, n_depth, d_features]} -- feature part 1

            Returns:
                output {Tensor, shape [batch, n_head * n_vchannel, n_depth, d_features]} -- output
                attn {Tensor, shape [n_head * batch, n_depth * n_blocks, n_depth * n_blocks]} -- self attention
        '''
        d_features, n_head, n_channel, n_vchannel, n_depth, d_block = \
            self.d_features, self.n_head, self.n_channel, self.n_vchannel, self.n_depth, self.d_block

        # let d_features divisible by d_block
        n_blocks = int(float(d_features) / d_block)  # d_features = n_blocks * d_block

        batch_size, _, _ = feature_map.size()

        query = self.query(feature_map)  # shape [batch, n_depth, n_head * n_channel * d_features]
        query = query.view(batch_size, n_depth, n_head, n_channel,
                           d_features)  # [batch, n_depth, n_head, n_channel, d_features]
        query = query.permute(2, 0, 1, 4, 3).contiguous()  # [n_head, batch, n_depth, d_features, n_channel]
        query = query.view(n_head * batch_size, n_depth, n_blocks, d_block,
                           n_channel)  # [n_head, batch, n_depth, n_blocks, d_block, n_channel]
        query = query.view(n_head * batch_size, n_depth * n_blocks,
                           n_channel * d_block)  # [n_head * batch, n_depth * n_blocks, n_channel * d_block]

        key = self.key(feature_map).view(batch_size, n_depth, n_head, n_channel,
                                         d_features)  # [batch, n_depth, n_head, n_channel, d_features]
        key = key.permute(2, 0, 1, 4, 3).contiguous()  # [n_head, batch, n_depth, d_features, n_channel]
        key = key.view(n_head * batch_size, n_depth * n_blocks,
                       n_channel * d_block)  # [n_head * batch, n_depth * n_blocks, n_channel * d_block]

        value = self.value(feature_map).view(batch_size, n_depth, n_head, n_vchannel,
                                             d_features)  # [batch, n_depth, n_head, n_vchannel, d_features]
        value = value.permute(2, 0, 1, 4, 3).contiguous()  # [n_head, batch, n_depth, d_features, n_vchannel]
        value = value.view(n_head * batch_size, n_depth * n_blocks,
                           n_vchannel * d_block)  # [n_head * batch, n_depth * n_blocks, n_vchannel * d_block]

        output, attn = self.attention(query, key,
                                      value)  # output: [n_head * batch, n_depth * n_blocks, n_vchannel * d_block],
        # attn: [n_head * batch, n_depth * n_blocks, n_depth * n_blocks]

        output = output.view(n_head, batch_size, n_depth, -1)  # [n_head, batch, n_depth, n_vchannel * d_features]
        output = output.permute(1, 2, 0, 3).contiguous().view(batch_size, n_depth, n_head * n_vchannel,
                                                              -1)  # [batch, n_depth, n_head * n_vchannel, d_features]
        output = output.transpose(1, 2)  # [batch, n_head * n_vchannel, n_depth, d_features]

        return output, attn


class ShuffleSelfAttentionLayerV2(nn.Module):
    def __init__(self, expansion_layer, n_depth, d_features, d_block, n_head, n_channel, n_vchannel, dropout, c_total=3,
                 mode='residual', use_bottleneck=True, d_bottleneck=None):
        super().__init__()

        assert n_depth > 1

        self.d_features = d_features
        d_features_new = torch.tensor(d_features, dtype=torch.float32)
        d_features_new = torch.ceil(d_features_new / d_block) * d_block
        self.d_features_new = d_features_new.int()

        self.c_total = c_total
        self.n_depth = n_depth
        self.mode = mode
        self.use_bottleneck = use_bottleneck

        self.index = feature_shuffle_index(d_features, depth=n_depth)
        self.index = torch.tensor(self.index)

        self.shuffle_slf_attn = ShuffleSelfAttentionV2(expansion_layer, n_depth, d_features, n_head, n_channel,
                                                       n_vchannel, d_block, c_total)

        self.layer_norm = nn.LayerNorm(d_features)
        self.dropout = nn.Dropout(dropout)

        self.conv2d = nn.Conv2d(n_head * n_vchannel, 1, kernel_size=(1, 1), bias=False)  # or use fc
        nn.init.xavier_normal(self.conv2d.weight)

        self.conv1d = nn.Conv1d(n_depth, 1, kernel_size=1, bias=False)
        nn.init.xavier_normal(self.conv1d.weight)

        self.avgpool = nn.AvgPool2d((n_depth, 1), stride=1)
        self.maxpool = nn.MaxPool2d((n_depth, 1), stride=1)

        if use_bottleneck:
            self.bottleneck = ShuffleBottleneckLayerV2(n_depth, d_features, mode, d_hid=d_bottleneck,
                                                       c_total=c_total + 2)

    def forward(self, features):
        '''
            Arguments:
                features {Tensor, shape: [batch, ?, d_features]}

        '''
        features = F.pad(features, (0, self.d_features_new - self.d_features), value=0)  # [batch, (?,) d_features_new]

        if features.dim() == 2:
            # The very first input layer, feature shape: [batch, d_features_new]
            features = features.view([-1, 1, self.d_features_new])  # [batch, 1, d_features_new]
            if self.mode == 'residual':
                paddings = torch.zeros_like(features).repeat([1, 2, 1])
                features = torch.cat((features, paddings), dim=1)  # [batch, 3, d_features_new]
                feature_map = features[:, :, self.index]  # [batch, 3, n_depth * d_features_new]
                feature_map = feature_map.view(
                    [-1, 3, self.n_depth, self.d_features])  # [batch, 3, n_depth, d_features_new]
            elif self.mode == 'dense':
                feature_map = features[:, :, self.index]  # [batch, 1, n_depth * d_features_new]
                feature_map = feature_map.view(
                    [-1, 1, self.n_depth, self.d_features])  # [batch, 1, n_depth, d_features_new]

        else:
            # feature shape: [batch, ?, d_features]
            feature_map = features[:, :, self.index]  # [batch, ?, n_depth * d_features_new]
            if self.mode == 'residual':
                feature_map = feature_map.view(
                    [-1, 3, self.n_depth, self.d_features])  # [batch, 3, n_depth, d_features_new]
            elif self.mode == 'dense':
                feature_map = feature_map.view(
                    [-1, self.c_total, self.n_depth, self.d_features])  # [batch, c_total, n_depth, d_features_new]

        residual = features  # [batch, ?, d_features_new]

        output, attn = self.shuffle_slf_attn(
            feature_map)  # [batch, n_head * n_vchannel, n_depth, d_features_new]
        output = self.conv2d(output)  # [batch, 1, n_depth, d_features_new]
        output = output.view([-1, self.n_depth, self.d_features_new])  # [batch, n_depth, d_features_new]

        avg_pooling = self.avgpool(output)  # [batch, 1, d_features_new]
        max_pooling = self.maxpool(output)  # [batch, 1, d_features_new]

        if self.mode == 'residual':
            # How to replace conv1d?
            conv = self.conv1d(output)  # [batch, 1, d_features_new]
            output = torch.cat([conv, avg_pooling, max_pooling], dim=1)  # [batch, 3, d_features_new]
            output = self.dropout(output)
            output = self.layer_norm(output + residual)

        if self.mode == 'dense':
            output = torch.cat([residual, avg_pooling, max_pooling], dim=1)  # [batch, c_total+2, d_features_new]
            output = self.dropout(output)
            output = self.layer_norm(output)

        if self.use_bottleneck:
            output = self.bottleneck(output)  # [batch, ?, d_features_new]

        return output, attn
